(function() {

    var LAST_ID = 0;

    var PieceStates = {
        Lay: 0,
        Snapped: 1,
        Up: 2,
        Tapped: 3
    };

    var Piece = function(opts) {
        var self = this;

        self.id = LAST_ID++;
        self.z = opts.zIndex || 0;
        self.image = opts.image;
        self.mask = opts.mask;
        self.targetX = opts.x;
        self.targetY = opts.y;
        self.parentNode = opts.parentNode;
        self.onPickUp = opts.onPickUp;
        self.onDrop = opts.onDrop;
        self.x = self.targetX;
        self.y = self.targetY;
        self.width = opts.width || self.image.naturalWidth || self.image.width;
        self.height = opts.height || self.image.naturalHeight || self.image.height;
        self.rotation = 0;
        self.state = PieceStates.Snapped;
        self.isRotating = false;
        self.snapPoint = null;

        var offset      = self.offset       = 50;             // Offset of the image, x-px padding
        var offsetHalf  = self.offsetHalf   = offset / 2;     // half of the above offset

        var node = document.createElement("canvas");
        node.style.position = "absolute";
        node.style.left = self.targetX + "px";
        node.style.top = self.targetY + "px";
        node.width = (self.width + offset) * Game.Utils.PixelRatio;
        node.height = (self.height + offset) * Game.Utils.PixelRatio;
        //node.setAttribute("width", node.width);
        //node.setAttribute("height", node.height);
        node.style.width = (self.width + offset) + "px";
        node.style.height = (self.height + offset) + "px";
        self.parentNode.appendChild(node);

        var context = node.getContext("2d");

        var rotationCallback = null;

        self.pickup = function() {
            if (self.isRotating) { return; }

            self.state = PieceStates.Up;
            if (self.snapPoint != null) {
                self.snapPoint.piece = null;
                self.snapPoint = null;
            }
            node.setAttribute("data-state", "up");

            if (self.onPickUp) {
                self.onPickUp(self);
            }

            //node.style.webkitTransform = "translate3d(0,0,0)";
            redraw();
            Game.Utils.ForceDOMRedraw(node);
        };

        self.tap = function() {
            self.state = PieceStates.Tapped;
            if (self.snapPoint != null) {
                self.snapPoint.piece = null;
                self.snapPoint = null;
            }
            node.setAttribute("data-state", "up");

            if (self.onPickUp) {
                self.onPickUp(self);
            }

            //node.style.webkitTransform = "translate3d(0,0,0)";
            redraw();
            Game.Utils.ForceDOMRedraw(node);
        }

        /**
         *
         * @param {Object} [snapPoint]
         */
        self.drop = function(snapPoint) {
            self.state = PieceStates.Lay;
            node.setAttribute("data-state", "");

            if (snapPoint) {
                self.snapPoint = snapPoint;
                self.snapPoint.piece = self;
                self.setPosition(self.snapPoint.x, self.snapPoint.y);
            }

            redraw();
            Game.Utils.ForceDOMRedraw(node);
        };


        /**
         * Rotate this Piece.
         * @param {function()} callback when done
         */
        self.rotate = function(callback) {
            if (self.isRotating) { return; }

            self.drop();
            self.isRotating = true;
            self.rotation += 90;

            if (self.rotation > 360) {
                self.rotation = 90;
            }

            node.style.webkitAnimation = "rotateTo" + self.rotation + " 0.75s cubic-bezier(.5, 0, .5, 1) forwards";
            rotationCallback = callback;
        };

        self.setRotation = function(deg) {
            self.rotation = deg % 360;
            node.style.webkitAnimation = "rotateTo" + self.rotation + " 1ms forwards";
            self.redraw();
        };

        function positionForElement(ele) {
            var x = 0,
                y = 0;
            while(true) {
                x += ele.offsetLeft; y += ele.offsetTop;
                if(ele.offsetParent === null) { break; }
                ele = ele.offsetParent;
            }
            return {x: x, y: y};
        }

        self.getPosition = function() {
            var ele = node;
            var x = 0,
                y = 0;
            while(true) {
                x += ele.offsetLeft; y += ele.offsetTop;
                if(ele.offsetParent === null) { break; }
                ele = ele.offsetParent;
            }
            return {x: x, y: y};
        };

        self.setPosition = function(x, y) {
            self.x = x;
            self.y = y;
            node.style.left = x + "px";
            node.style.top = y + "px";
        };


        self.canBePickedAt = function(x, y) {
            if (self.isRotating) { return false; }

            // simple bounding-box check, could be optimized with aabb
            /*if (x < self.x || x > self.x + self.width || y < self.y || y > self.y + self.height) {
                return false;
            }*/

            // checking for transparency
            var imgPosX = x - (self.x + (offsetHalf / Game.Utils.PixelRatio));
            var imgPosY = y - (self.y + (offsetHalf / Game.Utils.PixelRatio));

            // simple bounding-box check for relative image coords
            if (imgPosX < 0 || imgPosY < 0 || imgPosX > self.width || imgPosY > self.height) {
                return false;
            }

            // 0°:      x,  y
            // 90°:     y,  -x
            // 180°:   -x, -y
            // 270°:    -y, x
            if (!(self.rotation == 0 || self.rotation == 360)) {
                var tx = imgPosX - node.width * 0.5;
                var ty = imgPosY - node.height * 0.5;

                switch (self.rotation) {
                    case 90:
                        var tmp = tx;
                        tx = ty;
                        ty = -tmp;
                        break;
                    case 180:
                        tx = -tx;
                        ty = -ty;
                        break;
                    case 270:
                        var tmp = tx;
                        tx = -ty;
                        ty = tmp;
                        break;
                }

                imgPosX = tx + node.width * 0.5;
                imgPosY = ty + node.height * 0.5;

            }

            imgPosX *= Game.Utils.PixelRatio;
            imgPosY *= Game.Utils.PixelRatio;


            var imageData = context.getImageData(imgPosX + offsetHalf, imgPosY + offsetHalf, 1, 1);

            // when alpha is > 10 ( 255 max ) assume it's opaque
            // so let's handle this

            return (imageData.data[3] > 10);
        }


        self.equals = function(other) {
            if (!(other instanceof Piece)) return false;
            return self.targetX == other.targetX
                && self.targetY == other.targetY;
        }


        self.setZ = function(val) {
            //self.parentNode.removeChild(node);
            self.z = val;
            node.style.zIndex = self.z;
            //self.parentNode.appendChild(node);
            //node.style.webkitTransform = "translate3d(0,0,0)";
        };

        self.isAtTargetPosition = function() {
            return self.x == self.targetX
                && self.y == self.targetY
                && (self.rotation == 0 || self.rotation == 360);
        };


        // init z order
        self.setZ(self.z);

        self.delete = function() {
            self.parentNode.removeChild(node);
            node = null;
        };

        function forceDOMRedraw() {
            self.parentNode.removeChild(node);
            self.parentNode.appendChild(node);
        }

        function redraw() {
            //self.parentNode.removeChild(node);

            var shadowDistance,
                shadowBlur,
                shadowColor = "rgba(0, 0, 0, 0.3)";
            var lineWidth = 1;

            switch (self.state) {
                case PieceStates.Up:
                    shadowDistance = 20;
                    shadowBlur = 5;
                    break;
                case PieceStates.Snapped:
                    shadowDistance = 0;
                    shadowBlur = 0;
                    shadowColor = "rgba(0, 0, 0, 0)";
                    break;
                case PieceStates.Lay:
                    shadowDistance = 0;
                    shadowBlur = 5;
                    break;
                case PieceStates.Tapped:
                    shadowDistance = 0;
                    shadowBlur = 0;
                    break;
            }

            var shadowOffset = calculateShadowOffset(self.rotation, shadowDistance);

            // draw puzzle piece

            context.save();
            context.setTransform(1, 0, 0, 1, 0, 0);
            context.clearRect(0, 0, context.canvas.width, context.canvas.height);
            context.restore();
            var w = context.canvas.width;
            context.canvas.width = 1;
            context.canvas.width = w;

            context.save();
            context.shadowBlur = shadowBlur;
            context.shadowColor = shadowColor;
            context.shadowOffsetX = shadowOffset.x;
            context.shadowOffsetY = shadowOffset.y;
            context.drawImage(self.image, offsetHalf, offsetHalf, self.width * Game.Utils.PixelRatio, self.height * Game.Utils.PixelRatio);
            context.restore();

            if (self.state == PieceStates.Tapped) {
                context.save();
                context.fillStyle = "rgba(255, 255, 255, 0.125)";
                context.fillRect(offsetHalf, offsetHalf, self.width * Game.Utils.PixelRatio, self.height * Game.Utils.PixelRatio);
                context.restore();
            } else {
                drawTopEdge(self.rotation, lineWidth);
                drawLeftEdge(self.rotation, lineWidth);
                drawRightEdge(self.rotation, lineWidth);
                drawBottomEdge(self.rotation, lineWidth);
            }

            //self.parentNode.appendChild(node);

            /*if (self.id == 10) {
                context.save();
                context.fillStyle = "rgba(255, 0, 0, 0.25)";
                context.fillRect(Piece.OFFSET_HALF, Piece.OFFSET_HALF, self.width, self.height);
                context.restore();
            }*/
        }

        self.redraw = function() {
            redraw();
        };

        function lightColorForEdge() {
            if (self.isRotating || self.state == PieceStates.Up) {
                return "rgba(255, 255, 255, 0.125)";
            }
            return "rgba(255, 255, 255, 0.25)";
        }

        function darkColorForEdge() {
            if (self.isRotating || self.state == PieceStates.Up) {
                color = "rgba(255, 255, 255, 0.125)";
            }
            return "rgba(0, 0, 0, 0.5)";
        }

        function drawTopEdge(rot, lineWidth) {
            var color;
            if (self.isRotating || self.state == PieceStates.Up) {
                color = "rgba(255, 255, 255, 0.125)";
            } else {
                switch (rot) {
                    case 90:
                    case 180:
                        color = "rgba(0, 0, 0, 0.5)";
                        break;
                    default:
                        color = "rgba(255, 255, 255, 0.25)";
                        break;
                }
            }
            context.save();
            context.lineWidth = lineWidth;
            context.fillStyle = color;
            context.fillRect(offsetHalf, offsetHalf, self.width * Game.Utils.PixelRatio, lineWidth);
            context.restore();
        }

        function drawBottomEdge(rot, lineWidth) {
            var color;
            if (self.isRotating || self.state == PieceStates.Up) {
                color = "rgba(255, 255, 255, 0.125)";
            } else {
                switch (rot) {
                    case 90:
                    case 180:
                        color = "rgba(255, 255, 255, 0.25)";
                        break;
                    default:
                        color = "rgba(0, 0, 0, 0.5)";
                        break;
                }
            }
            context.save();
            context.lineWidth = lineWidth;
            context.fillStyle = color;
            context.fillRect(offsetHalf, offsetHalf + self.height * Game.Utils.PixelRatio, self.width * Game.Utils.PixelRatio, lineWidth);
            context.restore();
        }

        function drawLeftEdge(rot, lineWidth) {
            var color;
            if (self.isRotating || self.state == PieceStates.Up) {
                color = "rgba(255, 255, 255, 0.125)";
            } else {
                switch (rot) {
                    case 180:
                    case 270:
                        color = "rgba(0, 0, 0, 0.5)";
                        break;
                    default:
                        color = "rgba(255, 255, 255, 0.25)";
                        break;
                }
            }
            context.save();
            context.lineWidth = lineWidth;
            context.fillStyle = color;
            context.fillRect(offsetHalf, offsetHalf, lineWidth, self.height * Game.Utils.PixelRatio);
            context.restore();
        }

        function drawRightEdge(rot, lineWidth) {
            var color;
            if (self.isRotating || self.state == PieceStates.Up) {
                color = "rgba(255, 255, 255, 0.125)";
            } else {
                switch (rot) {
                    case 180:
                    case 270:
                        color = "rgba(255, 255, 255, 0.25)";
                        break;
                    default:
                        color = "rgba(0, 0, 0, 0.5)";
                        break;
                }
            }
            context.save();
            context.lineWidth = lineWidth;
            context.fillStyle = color;
            context.fillRect(offsetHalf + self.width * Game.Utils.PixelRatio - lineWidth, offsetHalf, lineWidth, self.height * Game.Utils.PixelRatio);
            context.restore();
        }


        function calculateShadowOffset(rot, distance) {
            // 0°:      x,  y
            // 90°:     y,  -x
            // 180°:   -x, -y
            // 270°:    -y, x
            var offset = { x: 0, y: 0};
            switch (rot) {
                case 90:
                    offset.x = distance;
                    offset.y = -distance;
                    break;
                case 180:
                    offset.x = -distance;
                    offset.y = -distance;
                    break;
                case 270:
                    offset.x = -distance;
                    offset.y = distance;
                    break;
                default:
                    offset.x = distance;
                    offset.y = distance;
                    break;
            }
            return offset;
        };

        // first draw
        redraw();

        node.addEventListener("webkitAnimationEnd", function() {
            self.isRotating = false;
            redraw();
            if (rotationCallback) { rotationCallback(); }
        }, false);

    }

    Piece.OFFSET = 50;
    Piece.OFFSET_HALF = 25;

    // Export
   window.Game.Piece = Piece;

})();